#include "format.h"
#include "zutil.h"
#include "search.h"
#include "searcher.h"

#include "search_bmsimple.h"
#include "search_bmsimple_multi.h"
#include "search_bmmultichar.h"
#include "search_bmmultichar_multi.h"
#include "search_bmblocks.h"
#include "search_bmblocks_multi.h"

#include "search_bm_onthefly.h" 
#include "search_bom_onthefly.h"
#include "search_kmp_onthefly.h"
#include "search_sbom_onthefly.h"
#include "search_ac_onthefly.h"



#include "lzgrep.h"
#include "blocks.h"
#include "util.h"
#include <errno.h>
#include <stdio.h>

static CompressFormat format;
static gzFile *gzSearchFile = NULL;

unsigned char SBUFF[DOBLE_BUF_SIZE+16];
//FILE *voc;

struct algorithm {
	char *name;
	int is_multipattern;
	void (*LZW_preproc)(void);
	int  (*LZW_search)(void);
	void (*LZ77_preproc)(void);
	int  (*LZ77_search)(void);
};

enum { 
	D_BM_onthefly,
	D_KMP_onthefly,
	D_BOM_onthefly,
	D_WM_onthefly,
	D_AC_onthefly,
	D_SBOM_onthefly,
	BM_simple,
	BM_simple_opt,
	BM_simple_lf,
	BM_multichar,
	BM_blocks,
	BM_blocks_opt,
	BM_blocks_2j,
	BM_simple_multi,
	BM_simple_multi_2,
	BM_simple_multi_3,
	BM_simple_multi_4,
      BM_multichar_multi,
	BM_multichar_multi_2,
	BM_blocks_multi,
	null_search
};
struct algorithm AlgorithmList[] = {
      {"D-BM", 0, preproc_bm_onthefly,search_bm_onthefly,preproc_bm_onthefly,search_bm_onthefly},
      {"D-KMP", 0, preproc_kmp_onthefly,search_kmp_onthefly,preproc_kmp_onthefly, search_kmp_onthefly},
      {"D-BOM", 0, preproc_bom_onthefly,search_bom_onthefly,preproc_bom_onthefly,search_bom_onthefly},
      {"D-WM", 1,preproc_sbom_onthefly,search_sbom_onthefly, preproc_sbom_onthefly,search_sbom_onthefly}, // OJO WM not yet implemented, acts as SBOM
      {"D-AC", 1, preproc_ac_onthefly,search_ac_onthefly,preproc_ac_onthefly,search_ac_onthefly},
      {"D-SBOM", 1,preproc_sbom_onthefly,search_sbom_onthefly, preproc_sbom_onthefly,search_sbom_onthefly},
      {"BM-simple-basic", 0, preproc_bmsimple,search_bmsimple_1,NULL,NULL},
	{"BM-simple", 0, preproc_bmsimple,search_bmsimple_2,NULL,NULL},
	{"BM-simple-lf", 0, preproc_bmsimple,search_bmsimple_3,NULL,NULL},
      {"BM-multichar", 0, preproc_bmmultichar,search_bmmultichar,NULL,NULL},
	{"BM-blocks-basic", 0, preproc_bmblocks_1,search_bmblocks_1,NULL,NULL},
	{"BM-blocks", 0, preproc_bmblocks_2,search_bmblocks_2,NULL,NULL},
	{"BM-blocks-2J", 0, preproc_bmblocks_3,search_bmblocks_3,NULL,NULL},
	{"BM-simple-multi-2", 1, preproc_bmsimple_multi_1,search_bmsimple_multi_1,NULL,NULL},
	{"BM-simple-multi-3", 1, preproc_bmsimple_multi_2,search_bmsimple_multi_2,NULL,NULL},
	{"BM-simple-multi-4", 1, preproc_bmsimple_multi_3,search_bmsimple_multi_3,NULL,NULL},
	{"BM-simple-multi", 1, preproc_bmsimple_multi_4,search_bmsimple_multi_4,NULL,NULL},
      {"BM-multichar-multi-2", 1, preproc_bmmultichar_multi_1,search_bmmultichar_multi_1,NULL},
	{"BM-multichar-multi", 1, preproc_bmmultichar_multi_2,search_bmmultichar_multi_2,NULL},
	{"BM-blocks-multi", 1, preproc_bmblocks_multi,search_bmblocks_multi,NULL,NULL},
	{NULL, 0, NULL, NULL, NULL, NULL}
};
int indexDefaultAlg=null_search;
	
/* Set algorithm to search */
void set_default_algorithm(char *algname) {
	int i=0;

      while ((AlgorithmList[i].name != NULL)	&&
               strcmp(AlgorithmList[i].name,algname)) 
		   i++;
	if (AlgorithmList[i].name == NULL) {
		fprintf(stderr,"%s: invalid algorithm argument\n",progname);
		exit(2);
	}

	indexDefaultAlg = i;
}

/* Gets the index of the best algorithm */
static int get_best_algorithm(CompressFormat format) {
	int index;

	/* LZW format */
        if (format == LZW_format) {
	    /* Monopattern */
	    if (PatternList.count == 1)
                  index = BM_simple_opt;
	    /* Multipattern */
	    else if (PatternList.count <= 10)
		  index = BM_simple_multi_4;
	    else if (PatternList.count <= 100)
		  index = D_WM_onthefly;
	    else 
		  index = D_SBOM_onthefly;
        }
	/* LZ77 format */
	else {
	    /* Monopattern */
	    if (PatternList.count == 1)
                  index = D_BOM_onthefly;
	    /* Multipattern */
	    if (PatternList.count <= 100)
                  index = D_WM_onthefly;
	    else 
		  index = D_SBOM_onthefly;

	}

	return(index-1);
}

/* Selects algorithm to search
 * and do the search
 * Returns 0 ==> match, 1==> no match, 2==>troubles */

int search() {

      int n;
	int i;
	void (*fpreproc)(void);
	int  (*fsearch)(void);

	format = get_format();
       
	gzSearchFile = NULL;
	if (format == LZW_format) {
		 if (!codeopen()) return(2);
	}

	if (format == LZ77_format) {
		gzSearchFile = gzopenbuf(filenameSearch,fdSearch,inbuf,insize);
		if (!gzSearchFile) {
		        fprintf(stderr, "%s: %s: Corrupt file\n",
		            progname,filenameSearch);
			return(2);
		} else fprintf(stderr, "%s: %s:  file LZ77 \n",
		            progname,filenameSearch);
	}

	if (format == unsupported_format) {
		fprintf(stderr,
		  "%s: %s: invalid or unsupported compression format\n",
		  progname,filenameSearch);
		return(2);
	}
        
	i = indexDefaultAlg;
	if ((!AlgorithmList[i].is_multipattern) && (PatternList.count > 1)) {
		fpreproc = NULL;
		fsearch = NULL;
	} else if (format == LZW_format) {
		fpreproc = AlgorithmList[i].LZW_preproc;
		fsearch = AlgorithmList[i].LZW_search;
	} else {
		//fprintf(stderr, "\nVai utilizar o algoritmo %s, %d \n", AlgorithmList[i].name, i);
		fpreproc = AlgorithmList[i].LZ77_preproc;
		fsearch = AlgorithmList[i].LZ77_search;
	}

	if (fpreproc == NULL || fsearch == NULL) {
		i = get_best_algorithm(format);
	        if (format == LZW_format) {
		   fpreproc = AlgorithmList[i].LZW_preproc;
		   fsearch = AlgorithmList[i].LZW_search;
	        } else {
		      // fprintf(stderr, "\nVai utilizar o algoritmo %s, %d \n", AlgorithmList[i].name, i);
		   fpreproc = AlgorithmList[i].LZ77_preproc;
		   fsearch = AlgorithmList[i].LZ77_search;
	        }
	}

    if (PatternList.count)   {
	fpreproc();
	//fprintf(stderr, "%s: %s: Pasa preproceso vai a buscar \n",
		           // progname,filenameSearch);
	//voc = fopen ("matches", "w");           
	n = fsearch();
	if (optListFiles==1) {
		if (n) printf("%s\n",filenameSearch);
	} else if (optListFiles==-1) {
		if (!n) printf("%s\n",filenameSearch);
	} else if (optCountMatches) {
		if (optPrintFilename) printf("%s:",filenameSearch);
		printf("Number of matches [%d]\n",n);
	}
    }
	if (gzSearchFile) gzclose(gzSearchFile);
	//fclose(voc);
	if (n < 0) return(2);
	else if (n > 0) return(0);
	else return(1);

}

void OTF_cleanState(OTF_SearchState *OTF_State) {
	OTF_State->rpos = 0;
	OTF_State->tpos = 0;
	OTF_State->count = 0;
	OTF_State->ppos = 0;
	OTF_State->oldbuf = SBUFF+4;
	OTF_State->remold = SBUFF;
	OTF_State->buf = SBUFF+BUF_SIZE+12;
	OTF_State->remnew=SBUFF+BUF_SIZE+8;
	OTF_State->linenum = 1;
	OTF_State->excess=0;

}

int OTF_readUncompressed(OTF_SearchState *OTF_State) {
	int rbytes;
	unsigned char *buf = OTF_State->oldbuf;
	unsigned char *rembuf = OTF_State->remold;
	unsigned char remain[4],excess[4];
	
	register unsigned char ii;
	
	if (OTF_State->tpos >0)
	for (ii=0; ii<4;ii++)
	       {remain[ii]=OTF_State->buf[BUF_SIZE+(ii-4)];
	        excess[ii]=OTF_State->buf[BUF_SIZE+ii];}
	  
     
	
	OTF_State->oldbuf = OTF_State->buf;
	OTF_State->remold=OTF_State->remnew;
	OTF_State->buf = buf;
	OTF_State->remnew=rembuf;
	if (OTF_State->tpos >0)
	for(ii=0;ii<4;ii++)
	 {OTF_State->remnew[ii]=remain[ii];
	  OTF_State->buf[ii]=excess[ii];
     }

	if (format == LZW_format) rbytes = getbuffer(buf,BUF_SIZE);
	else if (OTF_State->tpos ==0)rbytes = gzread(gzSearchFile,buf,BUF_SIZE+4);
	     else rbytes = gzread(gzSearchFile,buf+4,BUF_SIZE);

	//if (rbytes > 0) OTF_State->rpos += (rbytes+ii);
	if (rbytes > BUF_SIZE) {OTF_State->rpos += BUF_SIZE; rbytes=BUF_SIZE;}
	else OTF_State->rpos += rbytes;
	//OTF_State->inicio = OTF_State->tpos;
	
	return(rbytes);
}



int OTF_reportMatch(OTF_SearchState *OTF_State) {
	int nlines;
	int maxlen, maxlen1, maxlen2;
	unsigned char  *p = &OTF_State->buf[OTF_State->tpos&BUF_MASKM];
	unsigned long ppos;
	int dif,i,len;
	int pos;
	int brk = 0;
	char sep;
	LineList  *list, *nel;
	int aux, head;

	 
	
    //if (OTF_charAt((OTF_State->tpos)-1)>=128)  return(0); 
     
    
    if (OTF_State->buf[(long)(OTF_State->tpos&BUF_MASKM)-1-OTF_State->excess]>=S_value)  return(0); 
	OTF_State->count++;
    //fprintf(voc,"%d\n", OTF_State->tpos);
	//if (!indPrintMatches) goto OTF_afterMatch;
	/*if  ((OTF_State->count > 1)  && (optContextAfter || optContextBefore) && (OTF_State->ppos < OTF_State->tpos))
		printf("--\n");*/
	if (!optSilent){	
	//fprintf(stderr,"\n MATCH match num [%d]\n", OTF_State->count);	
	fprintf(stderr,"***MATCH num [%d] \n", OTF_State->count );
	dif = p - OTF_State->buf;
	if ( (dif >= 0) && (dif < BUF_SIZE)) {
		maxlen1 = dif;
		maxlen2 = BUF_SIZE;
	} else {
		maxlen1 = p - OTF_State->oldbuf;
		maxlen2 = 0;
	}

	dif = OTF_State->tpos - OTF_State->ppos;
	if (dif < 0) dif = 0;
	if (dif < (maxlen1+maxlen2)) {
		if (maxlen1 > dif) {
			maxlen1 = dif;
			maxlen2 = 0;
		} else {
			maxlen2 = dif - maxlen1;
		}
	}

	/* Collect lines before match */
	
	pos = OTF_State->tpos;
	
	/**** ETDC ****/
	

	

	len=maxlen1;
	pos=pos-maxlen1;
	
   if (len > 20)  {len = 20; pos=OTF_State->tpos-20;}
	
	//while ((OTF_charAt(pos) >=128) && (pos>0)) { pos++; len--;}
	while ((OTF_State->buf[pos&BUF_MASKM]>=S_value) && (pos>0)) { pos++; len--;}
	if (pos!=0) {pos++; len--;}
	
	nel = (LineList *)malloc(sizeof(LineList));
			//nel->buf = &OTF_charAt(pos);
			  nel->buf = &OTF_State->buf[pos&BUF_MASKM];
		        nel->pos = pos;
	if ((pos+len)<(OTF_State->tpos+long_patt[0])) len=len+((OTF_State->tpos+long_patt[0])-(pos+len));        
	nel->len = len;
	nel->pref = 1;
	//fprintf(stderr,"\n Antes [%d] len [%d]", OTF_charAt(pos+1),maxlen);
	


	//while (nel != NULL) {
	//	if (nel->pref) {
		//      sep = (nlines?'-':':');
		  //    if (optPrintFilename) 
			///      printf("%s%c",filenameSearch,sep);
		     // if (optLineNumber)
			//      printf("%d%c",OTF_State->linenum-nlines,sep);
		     // if (optByteOffset)
			   //   printf("%d%c",nel->pos,sep);

		//}

		
		imprime_texto_comprimido(nel->len,nel->buf,S_value);}
		
	//	free(nel);
	//}


/*OTF_afterMatch:

	    ppos = OTF_State->tpos;
		len=0;
	    p = &OTF_charAt(ppos);
		if ((p>=OTF_State->oldbuf)  && (p < OTF_State->oldbuf+BUF_SIZE))
			maxlen = OTF_State->oldbuf+BUF_SIZE - p;
		else
		    maxlen = OTF_State->rpos - ppos;
		
		        
*/		
		 //fprintf(stderr,"\n Despois de [%d] len [%d]", OTF_charAt(ppos),maxlen);        
	//	  while((len < maxlen-5) && (OTF_charAt(pos) >=128)) {      
		//while((len < maxlen) && (OTF_charAt(ppos)!=NEWLINE)) {
	//		len++;
	//		ppos++;
	//	}
		//fprintf(stderr,"\n Entra apuntando a %d len %d", OTF_charAt(p),len);        
		//if (OTF_charAt(ppos)==NEWLINE) {
		 
		     //imprime_texto_comprimido(len+1,p);
		
	return(0);

}

int OTF_reportMatchM(OTF_SearchState *OTF_State,int found,int inicio) {
	int nlines;
	int maxlen, maxlen1, maxlen2;
	unsigned char  *p = &OTF_State->buf[(OTF_State->tpos&BUF_MASKM)];
	unsigned long ppos;
	int dif,i,len;
	int pos;
	int brk = 0;
	char sep;
	LineList  *list, *nel;
	int aux, head,aumenta;
	
    
    
    //if (OTF_charAt(OTF_State->tpos-found)>=128) return(0);
    if (OTF_State->buf[(long)(OTF_State->tpos-inicio)-found]>=S_value) return(0);
     
    OTF_State->count++;
		//fprintf(voc,"%d\n", OTF_State->tpos-found+1);
	//if (!indPrintMatches) goto OTF_afterMatch;
	/*if  ((OTF_State->count > 1)  && (optContextAfter || optContextBefore) && (OTF_State->ppos < OTF_State->tpos))
		printf("--\n");*/
	if (!optSilent){	
	//fprintf(stderr,"\n MATCH match num [%d] en posicin [%d] Comprueba [%d]\n", OTF_State->count, OTF_State->tpos,OTF_charAt((OTF_State->tpos)-(found)));	
	fprintf(stderr,"***MATCH num [%d] \n", OTF_State->count );	
	dif = p  - OTF_State->buf -(found-1);
	aumenta=0;
	pos = OTF_State->tpos-(found-1);
	if (pos==0) 
	aumenta=found-1;
	if ( (dif >= 0) && (dif < BUF_SIZE)) {
		maxlen1 = dif;
	//	maxlen2 = BUF_SIZE;
	} else {
		maxlen1 = p - OTF_State->oldbuf;
	//	maxlen2 = 0;
	}

	dif = (pos -(found-1)) - OTF_State->ppos;
	if (dif < 0) dif = 0;
	
	/*if (dif < (maxlen1+maxlen2)) {
		if (maxlen1 > dif) {
			maxlen1 = dif;
			maxlen2 = 0;
		} else {
			maxlen2 = dif - maxlen1;
		}
	}*/

	/* Collect lines before match */
	

	
	/**** ETDC ****/
	

	

len=maxlen1;
	pos=pos-maxlen1;
	
   if (len > 20)  {len = 20; pos=OTF_State->tpos-20;}
	
	//while ((OTF_charAt(pos) >=128) && (pos>0)) { pos++; len--;}
	while ((OTF_State->buf[pos&BUF_MASKM]>=S_value) && (pos>0)) { pos++; len--;}
	if (pos!=0) {pos++; len--;}
	
	nel = (LineList *)malloc(sizeof(LineList));
			//nel->buf = &OTF_charAt(pos);
			  nel->buf = &OTF_State->buf[pos&BUF_MASKM];
		        nel->pos = pos;
	if ((pos+len)<(OTF_State->tpos+long_patt[0])) len=len+((OTF_State->tpos+long_patt[0])-(pos+len));        
	nel->len = len;
	nel->pref = 1;
		
		imprime_texto_comprimido(nel->len,nel->buf,S_value);}
		
	//	free(nel);
	//}


/*OTF_afterMatch:

	    ppos = OTF_State->tpos;
		len=0;
	    p = &OTF_charAt(ppos);
		if ((p>=OTF_State->oldbuf)  && (p < OTF_State->oldbuf+BUF_SIZE))
			maxlen = OTF_State->oldbuf+BUF_SIZE - p;
		else
		    maxlen = OTF_State->rpos - ppos;
		
		        
*/		
		 //fprintf(stderr,"\n Despois de [%d] len [%d]", OTF_charAt(ppos),maxlen);        
	//	  while((len < maxlen-5) && (OTF_charAt(pos) >=128)) {      
		//while((len < maxlen) && (OTF_charAt(ppos)!=NEWLINE)) {
	//		len++;
	//		ppos++;
	//	}
		//fprintf(stderr,"\n Entra apuntando a %d len %d", OTF_charAt(p),len);        
		//if (OTF_charAt(ppos)==NEWLINE) {
		 
		     //imprime_texto_comprimido(len+1,p);
		
	return(0);

}


void OTF_shift2(OTF_SearchState *OTF_State,int shift) {
   int i,pos;
   for (i=0,pos=OTF_State->tpos;i<shift;pos++,i++) 
	   if (OTF_charAt(pos) == NEWLINE) OTF_State->linenum++;
   OTF_State->tpos = pos;
}
